#!/usr/bin/env python3

"""
Helper script when RETURNN is run in server mode.
Will record some audio from microphone via ``sox``,
and then send it to a running RETURNN server instance.
"""

import sys
import os
from argparse import ArgumentParser
from subprocess import check_call
import better_exchook  # noqa

better_exchook.install()
arg_parser = ArgumentParser()
arg_parser.add_argument("--input_type", default="audio", help="audio or text. default is audio")
arg_parser.add_argument("--input_file", help="existing file")
arg_parser.add_argument("--input_text", help="some text")
arg_parser.add_argument("--ssh_host")
arg_parser.add_argument("--http_host", default="http://localhost:12380")
args = arg_parser.parse_args()

if args.input_file:
    assert os.path.exists(args.input_file)
    fn = args.input_file

elif args.input_type == "text" or args.input_text:
    if args.input_text:
        text = args.input_text
    else:
        text = input("Please insert some text: ")
    print("Text:", text)
    fn = "/tmp/input.txt"
    with open(fn, "w") as f:
        f.write("%s\n" % text)

elif args.input_type == "audio":
    # print("Noise profile")
    # check_call(["sox", "-n", "-d", "trim", "0", "1.5", "noiseprof", "/tmp/speech.noise-profile"])

    print("Using Sox for recording. Speak something, then pause 3 sec.")
    fn = "/tmp/rec.wav"
    if os.path.exists(fn):
        os.remove(fn)
    check_call(
        [
            "sox",
            "-d",
            "-c",
            "1",
            "-r",
            "16k",
            "-b",
            "16",
            fn,
            # "noisered", "/tmp/speech.noise-profile", "0.3",
            "silence",
            "1",
            "0.1",
            "0.1%",
            "1",
            "2.0",
            "1%",
            "norm",
        ],
        stdin=sys.stdin,
    )
    assert os.path.exists(fn)

else:
    raise Exception("invalid input type %r" % args.input_type)


def get_curl_cmd():
    """
    :rtype: list[str]
    """
    return ["curl", "-F", "file=@%s" % fn, args.http_host]


if args.ssh_host:
    print("copy to remote machine", args.ssh_host)
    check_call(["rsync", fn, "%s:/tmp" % args.ssh_host])
    fn = "/tmp/%s" % os.path.basename(fn)

    check_call(["ssh", args.ssh_host] + get_curl_cmd())

else:
    check_call(get_curl_cmd())



"""
This is a Sprint interface implementation.
See SprintInterface.py for another Sprint interface.
This Sprint interface is to be used for ExternSprintDataset, which should automatically use it.
"""

from __future__ import annotations

import sys
import os
import typing
from returnn.util import better_exchook
import returnn.util.task_system as task_system
from returnn.util.task_system import Pickler
from returnn.util.basic import to_bool, unicode, BytesIO

# Start Sprint PythonSegmentOrder interface. {
# We use the PythonSegmentOrder just to get an estimate (upper limit) about the number of sequences.

segmentOrderList = None  # type: typing.Optional[typing.List[str]]


# Cannot change name, this need to stay like this for compatibility.
# noinspection PyPep8Naming,PyUnusedLocal
def getSegmentList(corpusName, segmentList, **kwargs):
    """
    Called by Sprint PythonSegmentOrder.
    Set python-segment-order = true in Sprint to use this.

    If this is used, this gets called really early.
    If it is used together with the Sprint PythonTrainer,
    it will get called way earlier before the init() below.
    It might also get called multiple times, e.g. if
    Sprint is in interactive mode to calc the seg count.
    This is optional. You can use the SprintInterface
    only for the PythonTrainer.

    :type corpusName: str
    :type segmentList: list[str]
    :rtype: list[str]
    :returns segment list. Can also be an iterator.
    """
    print("SprintExternInterface: getSegmentList(%r), num segments: %i" % (corpusName, len(segmentList)))
    global segmentOrderList
    segmentOrderList = segmentList
    # No shuffling here. We expect to do that via Sprint.
    return segmentList


# End Sprint PythonSegmentOrder interface. }


# Start Sprint PythonTrainer interface. {

isInitialized = False
InputDim = None  # type: typing.Optional[int]
OutputDim = None  # type: typing.Optional[int]


def exchook(exc_type, exc_obj, exc_tb):
    """
    Replacement for sys.excepthook.
    """
    if exc_type is KeyboardInterrupt:
        print("SprintExternInterface[pid %i]: KeyboardInterrupt" % (os.getpid(),))
        sys.exit(1)
    better_exchook.better_exchook(exc_type, exc_obj, exc_tb)


def init(**kwargs):
    """
    Called by Sprint when it initializes the PythonTrainer,
    or also for PythonControl.
    Set trainer = python-trainer in Sprint to enable PythonTrainer for the Sprint nn-trainer tool.
    Note that Sprint will call this, i.e. the trainer init lazily quite late,
    only once it sees the first data.

    :param kwargs: all passed to :func:`_init_python_trainer` or :func:`PythonControl.init`
    :rtype: None|PythonControl
    """
    sys.excepthook = exchook
    # This module can also be used for Sprint PythonControl, which will also call init().
    # We need to catch these cases.
    if "name" in kwargs and kwargs["name"] == "Sprint.PythonControl":
        return PythonControl.init(**kwargs)
    return _init_python_trainer(**kwargs)


def _parse_config_str(config_str):
    assert isinstance(config_str, (str, unicode))
    config_list = config_str.split(",")
    config = {key: value for (key, value) in [s.split(":", 1) for s in config_list if s]}
    return config


def _common_init(config):
    if (
        to_bool(config.get("EnableAutoNumpySharedMemPickling", False))
        and not task_system.SharedMemNumpyConfig["enabled"]
    ):
        task_system.SharedMemNumpyConfig["enabled"] = True
        print("SprintExternInterface[pid %i] EnableAutoNumpySharedMemPickling = True" % (os.getpid(),))


# Cannot change argument names because of compatibility, as these are passed as-is from Sprint.
# noinspection PyPep8Naming
def _init_python_trainer(inputDim, outputDim, config, targetMode, **kwargs):
    """
    :type inputDim: int
    :type outputDim: int
    :param str config: config string, passed by Sprint. assumed to be ","-separated
    :param str targetMode: "target-alignment" or "criterion-by-sprint" or so
    """
    print("SprintExternInterface[pid %i]: PythonTrainer init_PythonTrainer()" % (os.getpid(),))
    print("inputDim:", inputDim)
    print("outputDim:", outputDim)
    print("config:", config)
    print("targetMode:", targetMode)
    print("other args:", kwargs)

    global InputDim, OutputDim, isInitialized
    InputDim = inputDim
    OutputDim = outputDim
    isInitialized = True
    assert targetMode != "criterion-by-sprint"
    config = _parse_config_str(config)
    assert config["action"] == "ExternSprintDataset"
    _common_init(config)

    _init_global_sprint_dataset(input_dim=inputDim, output_dim=outputDim, config=config)


sprintDataset = None  # type: typing.Optional[ExternSprintDatasetSource]


def _init_global_sprint_dataset(input_dim, output_dim, config):
    global sprintDataset
    if sprintDataset:
        return
    num_segments = len(segmentOrderList) if segmentOrderList is not None else None
    sprintDataset = ExternSprintDatasetSource(
        c2p_fd=int(config["c2p_fd"]),
        p2c_fd=int(config["p2c_fd"]),
        input_dim=input_dim,
        output_dim=output_dim,
        num_segments=num_segments,
    )


# Name need to stay like this, for compatibility.
# noinspection PyShadowingBuiltins
def exit():
    """
    Called by Sprint, to signal that it is exiting.
    """
    print("SprintExternInterface: PythonTrainer exit()")
    assert isInitialized
    sprintDataset.close()


# Name need to stay like this, for compatibility.
# noinspection PyPep8Naming
def feedInput(features, weights=None, segmentName=None):
    """
    Called by Sprint.
    Unsupervised case.

    :param numpy.ndarray features:
    :param numpy.ndarray|None weights:
    :param str|None segmentName:
    """
    feedInputAndTarget(features=features, weights=weights, segmentName=segmentName)


# Name need to stay like this, for compatibility.
# noinspection PyPep8Naming
def feedInputAndTargetAlignment(features, targetAlignment, weights=None, segmentName=None):
    """
    :param numpy.ndarray features:
    :param numpy.ndarray targetAlignment:
    :param numpy.ndarray|None weights:
    :param str|None segmentName:
    """
    feedInputAndTarget(features=features, alignment=targetAlignment, weights=weights, segmentName=segmentName)


# Name need to stay like this, for compatibility.
# noinspection PyPep8Naming
def feedInputAndTargetSegmentOrth(features, targetSegmentOrth, weights=None, segmentName=None):
    """
    :param numpy.ndarray features:
    :param str targetSegmentOrth:
    :param numpy.ndarray|None weights:
    :param str|None segmentName:
    """
    feedInputAndTarget(features=features, orthography=targetSegmentOrth, weights=weights, segmentName=segmentName)


feedInputUnsupervised = feedInput


# Name/params need to stay like this, for compatibility.
# noinspection PyPep8Naming,PyUnusedLocal
def feedInputAndTarget(
    features,
    weights=None,
    segmentName=None,
    orthography=None,
    alignment=None,
    speaker_name=None,
    speaker_gender=None,
    **kwargs,
):
    """
    :param numpy.ndarray features:
    :param numpy.ndarray|None weights:
    :param str|None segmentName:
    :param str|None orthography:
    :param numpy.ndarray|None alignment:
    :param str|None speaker_name:
    :param str|None speaker_gender:
    """
    assert features.shape[0] == InputDim
    targets = {}
    if alignment is not None:
        targets["classes"] = alignment
    if orthography is not None:
        targets["orth"] = orthography
    sprintDataset.add_new_data(segment_name=segmentName, features=features, targets=targets)


# End Sprint PythonTrainer interface. }


# Start Sprint PythonControl interface. {


class PythonControl:
    """
    PythonControl, interface for Sprint.
    """

    instance = None

    @classmethod
    def init(cls, **kwargs):
        """
        Called by global init().

        :rtype: PythonControl
        """
        print("SprintExternInterface[pid %i]: PythonControl %s init %r" % (os.getpid(), __file__, kwargs))
        if cls.instance:
            return cls.instance
        cls.instance = cls(**kwargs)
        return cls.instance

    # Maybe other kwargs by Sprint.
    # noinspection PyUnusedLocal
    def __init__(self, config, **kwargs):
        self.config = _parse_config_str(config)
        _common_init(self.config)

    def init_processing(self, input_dim, output_dim, **kwargs):
        """
        Called by Sprint.

        :param int input_dim:
        :param int output_dim:
        :param kwargs: maybe others
        """
        print(
            "SprintExternInterface: PythonControl init_processing inputDim=%i, outputDim=%i, other:%r"
            % (input_dim, output_dim, kwargs)
        )
        _init_global_sprint_dataset(input_dim=input_dim, output_dim=output_dim, config=self.config)

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def process_segment(self, name, orthography, features, alignment, soft_alignment, speaker_name=None, **kwargs):
        """
        Called by Sprint.

        :param str name:
        :param str|None orthography:
        :param numpy.ndarray features:
        :param numpy.ndarray|None alignment:
        :param numpy.ndarray|None soft_alignment:
        :param str|None speaker_name:
        :param kwargs: maybe others
        """
        assert sprintDataset
        targets = {}
        if orthography is not None:
            targets["orth"] = orthography
        if speaker_name is not None:
            targets["speaker_name"] = speaker_name
        if alignment is not None:
            targets["classes"] = alignment
        elif soft_alignment is not None:
            # We expect a sparse soft-alignment in coordinate format (time, class-idx, weight [0,1]).
            assert isinstance(soft_alignment, tuple)
            assert len(soft_alignment) == 3
            # We encode: sparse-coo-format, ndim == 2.
            targets["classes[sparse:coo:2:0]"] = soft_alignment[0]
            targets["classes[sparse:coo:2:1]"] = soft_alignment[1]
            targets["classes[sparse:coo:2:2]"] = soft_alignment[2]
        sprintDataset.add_new_data(segment_name=name, features=features, targets=targets)

    # noinspection PyMethodMayBeStatic
    def exit(self, **kwargs):
        """
        Called by Sprint.

        :param kwargs:
        """
        print("SprintExternInterface: PythonControl exit %r" % kwargs)
        if sprintDataset:
            sprintDataset.close()


# End Sprint PythonControl interface. }


class ExternSprintDatasetSource:
    """
    This will send data to ExternSprintDataset over a pipe.
    We expect that we are child process and the parent process has spawned us via ExternSprintDataset
    and is waiting for our data.
    """

    def __init__(self, c2p_fd, p2c_fd, input_dim, output_dim, num_segments):
        """
        :param int c2p_fd: child-to-parent file descriptor
        :param int p2c_fd: parent-to-child file descriptor
        :type input_dim: int
        :type output_dim: int
        :type num_segments: int | None
        :param num_segments: can be None if not known in advance
        """
        self.pipe_c2p = os.fdopen(c2p_fd, "wb")
        self.pipe_p2c = os.fdopen(p2c_fd, "rb")
        self._send("init", (input_dim, output_dim, num_segments))

    def _send(self, data_type, args=None):
        """
        :param str data_type:
        :param object args:
        """
        assert data_type is not None
        import struct

        stream = BytesIO()
        Pickler(stream).dump((data_type, args))
        raw_data = stream.getvalue()
        assert len(raw_data) > 0
        self.pipe_c2p.write(struct.pack("<i", len(raw_data)))
        self.pipe_c2p.write(raw_data)
        self.pipe_c2p.flush()

    def add_new_data(self, segment_name, features, targets):
        """
        :param str segment_name:
        :param numpy.ndarray features: 2D array, (feature,time)
        :param dict[str,numpy.ndarray] targets: each target is either 1D (time->idx) or 2D (time,class)
        """
        self._send("data", (segment_name, features, targets))

    def close(self):
        """
        Close pipe fds.
        """
        self._send("exit")
        self.pipe_c2p.close()
        self.pipe_p2c.close()


# End Sprint PythonControl interface. }



"""
This module covers the optimizer (SGD, Adam, etc) logic,
and model param update logic in general.
"""

from __future__ import annotations

import torch
import typing
from typing import Any, Set, Dict

from returnn.log import log
from returnn.util.basic import RefIdEq

_OptimizerClassesDictInitialized = False
_OptimizerClassesDict = {}


def _init_optimizer_classes_dict():
    """
    Initializes a global dictionary with all optimizers available in PyTorch.
    """
    global _OptimizerClassesDictInitialized
    if _OptimizerClassesDictInitialized:
        return
    _OptimizerClassesDictInitialized = True
    for name, cls in list(vars(torch.optim).items()):
        assert isinstance(name, str)
        # Check if cls is a valid subclass of torch.optim.Optimizer
        if not isinstance(cls, type) or not issubclass(cls, torch.optim.Optimizer):
            continue
        assert name not in _OptimizerClassesDict
        _OptimizerClassesDict[name.lower()] = cls


def get_optimizer_class(class_name):
    """
    :param str|()->torch.optim.Optimizer|type[torch.optim.Optimizer] class_name:
        Optimizer data, e.g. "adam", torch.optim.Adam...
    :return: Optimizer class
    :rtype: type[torch.optim.Optimizer]
    """
    _init_optimizer_classes_dict()
    if isinstance(class_name, type):
        assert issubclass(class_name, torch.optim.Optimizer)
    elif callable(class_name):
        class_name = class_name()
    else:
        assert isinstance(class_name, str)
        assert (
            class_name.lower() in _OptimizerClassesDict
        ), "%s not found in the available torch optimizers list: %s." % (
            class_name.lower(),
            ", ".join("'%s'" % key for key in _OptimizerClassesDict),
        )
        class_name = _OptimizerClassesDict[class_name.lower()]

    return class_name


def _get_class_init_kwargs(optim_class):
    """
    Obtains the keyword arguments of the class provided as parameter that the user can add to their optimizer.

    :param type[torch.optim.Optimizer] optim_class: Optimizer class.
    :return: Keyword arguments of the provided class.
    :rtype: List[str]
    """
    from returnn.util.basic import collect_class_init_kwargs

    optim_class_init_kwargs = collect_class_init_kwargs(optim_class)
    # We already provide params by default, remove it so that the user doesn't add it to the optimizer dict.
    optim_class_init_kwargs.remove("params")

    return optim_class_init_kwargs


class Updater(object):
    """
    Wraps a torch.optim.Optimizer, and extends it by some further functionality.
    """

    def __init__(self, config, network, initial_learning_rate=1.0):
        """
        :param returnn.config.Config config: config defining the training conditions.
        :param torch.nn.Module network: PyTorch Module defining the network.
        :param float initial_learning_rate:
        """
        self.config = config
        self.learning_rate = initial_learning_rate
        self.network = network
        self.optimizer = None  # type: typing.Optional[torch.optim.Optimizer]

    def set_learning_rate(self, value):
        """
        Updates the learning rate of the optimizer at each (sub)epoch.

        :param float value: New learning rate.
        """
        for param_group in self.optimizer.param_groups:
            param_group["lr"] = value

    def get_current_step_learning_rate(self):
        """
        Obtains an updated learning rate for the current training step inside a (sub)epoch.
        """
        pass

    def create_optimizer(self):
        """
        Creates an optimizer and stores it in self.optimizer.
        """
        optimizer_opts = self.config.typed_value("optimizer", None)
        if optimizer_opts is None:
            raise ValueError("config field 'optimizer' needs to be set explicitely for the Torch backend")
        self.optimizer = self._create_optimizer(optimizer_opts)

    def load_optimizer(self, filename):
        """
        Loads a torch.optim.Optimizer from disk and stores it in self.optimizer.

        :param str filename: File from which to load the optimizer state.
        """
        print("Load optimizer %s" % filename, file=log.v4)
        optimizer_state = torch.load(filename)
        self.optimizer.load_state_dict(optimizer_state)

    def save_optimizer(self, filename):
        """
        Saves the state of self.optimizer to a file.

        :param str filename: File in which to save the optimizer state.
        """
        print("Save optimizer under %s" % filename, file=log.v4)
        torch.save(self.optimizer.state_dict(), filename)

    def get_optimizer(self):
        """
        :return: Wrapped optimizer object.
        :rtype: torch.optim.Optimizer
        """
        return self.optimizer

    def _create_optimizer(self, optimizer_opts):
        """
        Returns a valid optimizer considering the dictionary given by the user in the config.

        :param dict[str]|str optimizer_opts: Optimizer configuration specified by the user.
            If it's a dict, it must contain "class" with the optimizer name or callable.
            If it's a str, it must be the optimizer name.
        :return: A valid optimizer.
        :rtype: torch.optim.Optimizer
        """
        lr = self.learning_rate

        # If the parameter is already a valid optimizer, return it without further processing
        if isinstance(optimizer_opts, torch.optim.Optimizer):
            return optimizer_opts
        elif callable(optimizer_opts):
            optimizer_opts: Dict[str, Any] = {"class": optimizer_opts}
        else:
            if not isinstance(optimizer_opts, dict):
                raise ValueError("'optimizer' must of type dict, callable or torch.optim.Optimizer instance.")
            if "class" not in optimizer_opts:
                raise ValueError("'class' field of 'optimizer' dict was not set (use e.g. 'SGD', 'Adam', ...)")
            optimizer_opts = optimizer_opts.copy()

        # Resolve the optimizer class
        optim_class_name = optimizer_opts.pop("class")
        optim_class = get_optimizer_class(optim_class_name)

        # Resolve the optimizer arguments
        opt_kwargs = optimizer_opts.copy()
        optim_class_init_kwargs = _get_class_init_kwargs(optim_class)
        # epsilon is named eps in torch.
        # If the user specified it as epsilon, parse it as eps for the optimizer
        if "eps" in optim_class_init_kwargs and "epsilon" in opt_kwargs:
            opt_kwargs["eps"] = opt_kwargs.pop("epsilon")
        if "learning_rate" in optimizer_opts:
            raise ValueError("'learning_rate' should be set outside of the 'optimizer' dict.")
        lr = lr * optimizer_opts.get("learning_rate_multiplier", 1.0)
        opt_kwargs["lr"] = lr

        param_groups = self._get_optimizer_param_groups(optim_class, opt_kwargs)
        optimizer = optim_class(param_groups, **opt_kwargs)
        print("Optimizer: %s" % optimizer, file=log.v1)
        assert isinstance(optimizer, torch.optim.Optimizer)

        return optimizer

    def _create_default_optimizer(self):
        """
        :return: SGD optimizer.
        :rtype: torch.optim.SGD
        """
        print("Create SGD optimizer (default).", file=log.v2)
        optimizer = torch.optim.SGD(self.network.parameters(), lr=self.learning_rate)

        return optimizer

    def _get_optimizer_param_groups(self, optim_class, optimizer_opts):
        """
        The weight_decay parameter from AdamW affects the weights of layers such as LayerNorm and Embedding.
        This function creates a blacklist of network modules and splits the optimizer groups in two:
        those who will receive weight decay, and those who won't receive it.
        The weight_decay parameter of the rest of the optimizers is L2 regularization.

        For further reading, see https://github.com/karpathy/minGPT/pull/24#issuecomment-679316025 and
        https://discuss.pytorch.org/t/weight-decay-in-the-optimizers-is-a-bad-idea-especially-with-batchnorm/16994.

        This code is based on https://github.com/karpathy/minGPT (MIT license):
        https://github.com/karpathy/minGPT/blob/3ed14b2cec0dfdad3f4b2831f2b4a86d11aef150/mingpt/model.py#L136.

        :param type[torch.optim.Optimizer] optim_class: Optimizer class.
        :param dict[str] optimizer_opts: Optimizer configuration specified by the user.
        :return: List of configurations for the different sets of parameters.
        :rtype: List[Dict[str]]
        """
        network_params = self.network.parameters()

        # By default insert the weight_decay constraints in the optimizer, as this is default PyTorch behavior.
        # If the user doesn't accept this, throw an error message.
        assert self.config.bool("decouple_constraints", True), (
            "L2/weight_decay constraints are decoupled in PyTorch, but "
            "decouple_constraints=False was explicitly specified in the config."
        )

        # Split in parameter groups only if decouple_constraints is set and the optimizer accepts weight_decay.
        cls_init_kwargs = _get_class_init_kwargs(optim_class)
        if "weight_decay" not in cls_init_kwargs:
            assert (
                "weight_decay" not in optimizer_opts
            ), "weight_decay not accepted by the chosen optimizer. Accepted values: %s" % ", ".join(
                "%s" % optim_name for optim_name in cls_init_kwargs
            )
            return [{"params": network_params}]

        weight_decay = optimizer_opts.get("weight_decay", 0.0)
        if not weight_decay:
            return [{"params": network_params}]

        # Distinguish between parameters with and without weight_decay/L2 regularization.
        # Parameters without weight decay: biases + LayerNorm/Embedding layers.
        wd_params = set()
        no_wd_params = set()
        blacklist_wd_modules = (torch.nn.LayerNorm, torch.nn.Embedding)
        # Tracker of visited parameters to only add each parameter once, in case two modules share common parameters.
        # We need the wrapper class RefIdEq because Parameters are compared by value and not by reference.
        visited_params: Set[RefIdEq[torch.nn.Parameter]] = set()
        for module_name, module in self.network.named_modules():
            module_name: str
            module: torch.nn.Module
            for param_name, param in module.named_parameters(recurse=False):
                param_name: str
                param: torch.nn.Parameter
                if RefIdEq(param) in visited_params:
                    continue
                visited_params.add(RefIdEq(param))
                full_param_name = "%s.%s" % (module_name, param_name) if module_name else param_name
                if param_name.endswith("bias"):
                    no_wd_params.add(full_param_name)
                elif param_name.endswith("weight") and isinstance(module, blacklist_wd_modules):
                    no_wd_params.add(full_param_name)
                else:
                    wd_params.add(full_param_name)

        param_dict = {pn: p for pn, p in self.network.named_parameters()}
        optim_groups = [
            {"params": [param_dict[pn] for pn in sorted(list(wd_params))], "weight_decay": weight_decay},
            {"params": [param_dict[pn] for pn in sorted(list(no_wd_params))], "weight_decay": 0.0},
        ]

        return optim_groups



"""
RETURNN frontend (returnn.frontend) tests
"""

from __future__ import annotations
import _setup_test_env  # noqa
from collections import OrderedDict
import returnn.frontend as rf
from returnn.tensor import Tensor, Dim, TensorDict, batch_dim
from rf_utils import run_model


def test_module_list():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
        }
    )

    class _Net(rf.Module):
        def __init__(self):
            super().__init__()
            self.base_dim = Dim(3, name="linear-out")
            dims = [self.base_dim + i for i in range(4)]
            in_dims = [in_dim] + dims[:-1]
            self.out_dim = dims[-1]
            self.ls = rf.ModuleList([rf.Linear(in_dim_, out_dim_) for in_dim_, out_dim_ in zip(in_dims, dims)])

        def __call__(self, out: Tensor) -> Tensor:
            """
            Forward
            """
            for layer in self.ls:
                out = layer(out)
            return out

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        out = model(extern_data["data"])
        out.mark_as_default_output(shape=(batch_dim, time_dim, model.out_dim))

    run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step)


def test_sequential_base_case():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
        }
    )

    class _Net(rf.Module):
        def __init__(self):
            super().__init__()
            dims = [Dim(1, name="feat1"), Dim(2, name="feat2"), Dim(3, name="feat3")]
            in_dims = [in_dim] + dims[:-1]
            self.out_dim = dims[-1]
            self.seq = rf.Sequential(rf.Linear(in_dim_, out_dim_) for in_dim_, out_dim_ in zip(in_dims, dims))

        def __call__(self, data: Tensor) -> Tensor:
            """
            Forward
            """
            seq = self.seq(data)
            return seq

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        out = model(extern_data["data"])
        out.mark_as_default_output(shape=(batch_dim, time_dim, model.out_dim))

    run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step)


def test_sequential_named_case():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
        }
    )

    class _Net(rf.Module):
        def __init__(self):
            super().__init__()
            dims = [Dim(1, name="feat1"), Dim(2, name="feat2"), Dim(3, name="feat3")]
            self.out_dim = dims[-1]
            x = OrderedDict()
            x["one"] = rf.Linear(in_dim, dims[0])
            x["two"] = rf.Linear(dims[0], dims[1])
            x["three"] = rf.Linear(dims[1], dims[2])
            self.seq = rf.Sequential(x)

        def __call__(self, data: Tensor) -> Tensor:
            """
            Forward
            """
            seq = self.seq(data)
            return seq

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        out = model(extern_data["data"])
        out.mark_as_default_output(shape=(batch_dim, time_dim, model.out_dim))

    run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step)


def test_parameter_list():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
        }
    )

    class _Net(rf.Module):
        def __init__(self):
            super().__init__()
            self.param_list = rf.ParameterList([rf.Parameter([in_dim]) for _ in range(3)])

        def __call__(self, data: Tensor) -> Tensor:
            """
            Forward
            """
            for param in self.param_list:
                data += param
            return data

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        out = model(extern_data["data"])
        out.mark_as_default_output(shape=(batch_dim, time_dim, in_dim))

    run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step)



"""
Here we encapsulate some common Horovod functions.

Note that you are supposed to be able to import this module even if Horovod is not installed.

The usage of this module / global context is also considered optional at this point.
Horovod is enabled <==> ``use_horovod`` is enabled in the config.

For relevant further config options, see the code of :class:`HorovodContext` below.
Most importantly:

* ``horovod_dataset_distribution``, recommended value ``"random_seed_offset"``, default value ``"shard"``
* ``horovod_reduce_type``, recommended value ``"param"``, default value ``"grad"``
* ``horovod_param_sync_step``, recommended value ``100``, default value ``1``
* ``horovod_param_sync_time_diff``, alternative to ``horovod_param_sync_step``, e.g. ``100.`` (secs), default ``None``

Also see :ref:`multi_gpu`.
Also see :mod:`TFDistributed`.
"""

import os
import socket
import typing
from returnn.config import Config


class HorovodContext:
    """
    This setups some helper functions.
    """

    def __init__(self, config):
        """
        :param Config config:
        """
        # noinspection PyUnresolvedReferences,PyPackageRequirements
        import horovod

        print("Horovod:", horovod.__version__, horovod.__file__)
        # noinspection PyUnresolvedReferences,PyPackageRequirements
        import horovod.tensorflow as hvd

        hvd.init()
        print(
            "Horovod initialized. Hostname %s, pid %i, rank %i / size %i, local rank %i / local size %i."
            % (socket.gethostname(), os.getpid(), hvd.rank(), hvd.size(), hvd.local_rank(), hvd.local_size())
        )
        self._config = config
        self._hvd_mod = hvd
        self._local_rank = hvd.local_rank()
        self._local_size = hvd.local_size()
        self._rank = hvd.rank()
        self._size = hvd.size()
        # Note: Rather than reading the config here, we will instead read it in the functions below on-the-fly.
        # This allows e.g. the pretraining or custom get_network to temporarily overwrite the behavior for some epoch.

    def should_sync_every_step(self):
        """
        :return: whether we should sync every step.
          This is both for the signal for more data, and also loss/error/score reduction.
        :rtype: bool
        """
        if self.is_dataset_distribution_random_seed_offset() and self.is_reduce_type_param():
            # In this setting, we don't need a sync every step.
            # Thus avoid it to have faster more async training.
            return False
        # By default, we should sync every step.
        return True

    def get_reduce_type(self):
        """
        :rtype: str
        """
        reduce_type = self._config.value("horovod_reduce_type", "grad")
        assert reduce_type in {"grad", "param"}
        return reduce_type

    def is_reduce_type_grad(self):
        """
        :rtype: bool
        """
        return self.get_reduce_type() == "grad"

    def is_reduce_type_param(self):
        """
        :rtype: bool
        """
        return self.get_reduce_type() == "param"

    def get_param_sync_time_diff(self):
        """
        :rtype: float|None
        """
        assert self.is_reduce_type_param()
        return self._config.float("horovod_param_sync_time_diff", None)

    def get_param_sync_step(self):
        """
        :rtype: int
        """
        assert self.is_reduce_type_param()
        return self._config.int("horovod_param_sync_step", 1)

    def get_dataset_distribution_type(self):
        """
        :rtype: str
        """
        dataset_distribution = self._config.value("horovod_dataset_distribution", "shard")
        assert dataset_distribution in {"shard", "random_seed_offset"}
        return dataset_distribution

    def is_dataset_distribution_shard(self):
        """
        :rtype: bool
        """
        return self.get_dataset_distribution_type() == "shard"

    def get_dataset_shard_batch_slice(self):
        """
        :rtype: slice
        """
        assert self.is_dataset_distribution_shard()
        return slice(self.rank(), None, self.size())

    def is_dataset_distribution_random_seed_offset(self):
        """
        :rtype: bool
        """
        return self.get_dataset_distribution_type() == "random_seed_offset"

    def rank(self):
        """
        :rtype: int
        """
        return self._rank

    def size(self):
        """
        :rtype: int
        """
        return self._size

    def local_rank(self):
        """
        :rtype: int
        """
        return self._local_rank

    def local_size(self):
        """
        :rtype: int
        """
        return self._local_size


_is_set_up = False
_ctx = None  # type: typing.Optional[HorovodContext]


def get_ctx(config=None):
    """
    :param Config|None config:
    :returns: the global context if Horovod is enabled, or None otherwise.
      If we did not setup the context yet, it will automatically create it.
    :rtype: HorovodContext|None
    """
    global _is_set_up, _ctx
    if _is_set_up:
        return _ctx
    if not config:
        from returnn.config import get_global_config

        config = get_global_config(raise_exception=False)
        if not config:
            return None
    _is_set_up = True
    if not config.is_true("use_horovod"):
        return None
    _ctx = HorovodContext(config=config)
    return _ctx



"""
copied from https://github.com/HawkAaron/warp-transducer/blob/master/tensorflow_binding/warprnnt_tensorflow/__init__.py
Changes: RNN-T -> RNA.
"""
import imp
import tensorflow as tf
from tensorflow.python.framework import ops

lib_file = imp.find_module('kernels', __path__)[1]
_warprna = tf.load_op_library(lib_file)


def rna_loss(log_probs, labels, input_lengths, label_lengths, blank_label=0):
    '''Computes the RNA loss between a sequence of activations and a
    ground truth labeling.
    Args:
        log_probs: A 4-D Tensor of floats.  The dimensions
                     should be (B, T, U, V), where B is the minibatch index,
                     T is the time index, U is the prediction network sequence
                     length, and V indexes over activations for each
                     symbol in the alphabet.
        labels: A 2-D Tensor of ints, a padded label sequences to make sure
                     labels for the minibatch are same length.
        input_lengths: A 1-D Tensor of ints, the number of time steps
                       for each sequence in the minibatch.
        label_lengths: A 1-D Tensor of ints, the length of each label
                       for each example in the minibatch.
        blank_label: int, the label value/index that the RNA
                     calculation should use as the blank label
    Returns:
        1-D float Tensor, the cost of each example in the minibatch
        (as negative log probabilities).
    '''
    loss, _ = _warprna.warp_rna(log_probs, labels, input_lengths,
                                label_lengths, blank_label)
    return loss


@ops.RegisterGradient("WarpRNA")
def _RNALossGrad(op, grad_loss, _):
    grad = op.outputs[1]
    # NOTE since here we are batch first, cannot use _BroadcastMul
    grad_loss = tf.reshape(grad_loss, (-1, 1, 1, 1))
    return [grad_loss * grad, None, None, None]


@ops.RegisterShape("WarpRNA")
def _RNALossShape(op):
    inputs_shape = op.inputs[0].get_shape().with_rank(4)
    batch_size = inputs_shape[0]
    return [batch_size, inputs_shape]



"""
Some helper utils.
"""


from __future__ import annotations
from typing import Optional, Union, Dict
import numpy
from returnn.tensor import Tensor, Dim, TensorDict, batch_dim


def tensor_dict_fill_random_numpy_(
    tensor_dict: TensorDict,
    *,
    rnd: Union[int, numpy.random.RandomState] = 42,
    dyn_dim_max_sizes: Optional[Dict[Dim, int]] = None,
    dyn_dim_min_sizes: Optional[Dict[Dim, int]] = None,
):
    """
    Random fill with NumPy arrays.

    :param tensor_dict:
    :param rnd:
    :param dyn_dim_max_sizes: you can specify max sizes for dim tags with dynamic sizes.
        The fill random code makes sure that there is at least one entry where we reach the max size,
        so that the dim value will be the max size.
    :param dyn_dim_min_sizes:
    """
    if not isinstance(rnd, numpy.random.RandomState):
        rnd = numpy.random.RandomState(rnd)
    for v in tensor_dict.data.values():
        tensor_fill_random_numpy_(v, rnd=rnd, dyn_dim_max_sizes=dyn_dim_max_sizes, dyn_dim_min_sizes=dyn_dim_min_sizes)


def tensor_fill_random_numpy_(
    x: Tensor,
    *,
    min_val: int = 0,
    max_val: Optional[int] = None,
    rnd: numpy.random.RandomState,
    dyn_dim_max_sizes: Optional[Dict[Dim, int]] = None,
    dyn_dim_min_sizes: Optional[Dict[Dim, int]] = None,
) -> bool:
    """fill. return whether sth was filled"""
    if dyn_dim_max_sizes is None:
        dyn_dim_max_sizes = {}
    if dyn_dim_min_sizes is None:
        dyn_dim_min_sizes = {}
    filled = False
    while True:
        have_unfilled = False
        filled_this_round = False

        for dim in x.dims:
            if dim.is_batch_dim() and not dim.dyn_size_ext:
                dim.dyn_size_ext = Tensor("batch", [], dtype="int32")
            if dim.is_dynamic() and not dim.dyn_size_ext:
                dim.dyn_size_ext = Tensor(dim.name or "time", dims=[batch_dim], dtype="int32")
            if not dim.dyn_size_ext:
                continue
            if tensor_fill_random_numpy_(
                dim.dyn_size_ext,
                min_val=dyn_dim_min_sizes.get(dim, 2),
                max_val=dyn_dim_max_sizes.get(dim, None),
                rnd=rnd,
                dyn_dim_max_sizes=dyn_dim_max_sizes,
            ):
                if dim in dyn_dim_max_sizes:
                    # Make sure at least one of the dyn sizes matches the max size.
                    i = rnd.randint(0, dim.dyn_size_ext.raw_tensor.size)
                    dim.dyn_size_ext.raw_tensor.flat[i] = dyn_dim_max_sizes[dim]
                    if dim in dyn_dim_min_sizes:
                        j = rnd.randint(0, dim.dyn_size_ext.raw_tensor.size - 1)
                        if j >= i:
                            j += 1
                        dim.dyn_size_ext.raw_tensor.flat[j] = dyn_dim_min_sizes[dim]
                elif dim in dyn_dim_min_sizes:
                    raise Exception(f"also define {dim} in dyn_dim_max_sizes, not just dyn_dim_min_sizes")
                filled = True
                filled_this_round = True
            if dim.dyn_size_ext.raw_tensor is None:
                have_unfilled = True
            elif not isinstance(dim.dyn_size_ext.raw_tensor, numpy.ndarray):
                have_unfilled = True

        if have_unfilled:
            assert filled_this_round, f"should have filled something, {x}"

        if not have_unfilled:
            break

    if x.raw_tensor is not None:
        if not isinstance(x.raw_tensor, numpy.ndarray):
            x.raw_tensor = None

    if x.raw_tensor is None:
        shape = [d.get_dim_value() for d in x.dims]
        if x.dtype.startswith("int"):
            if max_val is None:
                max_val = rnd.randint(5, 20)
            if x.sparse_dim and x.sparse_dim.dimension is not None:
                max_val = x.sparse_dim.dimension
            x.raw_tensor = rnd.randint(min_val, max_val, size=shape, dtype=x.dtype)
        elif x.dtype == "bool":
            x.raw_tensor = rnd.randint(0, 2, size=shape, dtype=x.dtype)
        elif x.dtype.startswith("float"):
            x.raw_tensor = rnd.normal(0.0, 1.0, size=shape).astype(x.dtype)
        elif x.dtype.startswith("complex"):
            real = rnd.normal(0.0, 1.0, size=shape)
            imag = rnd.normal(0.0, 1.0, size=shape)
            x.raw_tensor = (real + 1j * imag).astype(x.dtype)
        else:
            raise NotImplementedError(f"not implemented for {x} dtype {x.dtype}")
        filled = True

    assert isinstance(x.raw_tensor, numpy.ndarray)

    return filled
